/**
 * This program has been developed by students from the bachelor Computer Science at
 * Utrecht University within the Software Project course.
 * © Copyright Utrecht University (Department of Information and Computing Sciences)
 */

/* istanbul ignore file */
/* The comment above was added so the code coverage wouldn't count this file towards code coverage.
 * We do not test components/renderfunctions/styling files.
 * See testing plan for more details.*/
import React, { useCallback, useMemo } from 'react';
import { Handle, HandleType, NodeProps, Position } from 'reactflow';
import styles from './logicpill.module.scss';
import {
  AllLogicMap,
  EntityNodeAttributes,
  Handles,
  LogicData,
  LogicNodeAttributes,
  QueryGraphEdges,
  QueryGraphNodes,
  SchemaReactflowLogicNode,
  toHandleData,
  toHandleId,
} from '../../../model';
import { Input } from '@mui/material';
import { styleHandleMap } from '../../utils';
import {
  setQuerybuilderNodes,
  useAppDispatch,
  useQuerybuilderGraph,
  useQuerybuilderGraphology,
} from '@graphpolaris/shared/lib/data-access';
import { InputNode, InputNodeTypeTypes } from '../../../model/logic/general';
import { SerializedEdge, SerializedNode } from 'graphology-types';

/**
 * Component to render an entity flow element
 * @param param0 Data of the flow element.
 */
export default function LogicPill(node: SchemaReactflowLogicNode) {
  const dispatch = useAppDispatch();
  const data = node.data;
  const logic = data.logic;
  const output = data.logic.output;

  const graph = useQuerybuilderGraph();
  const graphology = useQuerybuilderGraphology();
  const connectionsToLeft = useMemo(() => graph.edges.filter((edge) => edge.target === node.id), [graph]);
  const connectionsToRight = useMemo(() => graph.edges.filter((edge) => edge.source === node.id), [graph]);

  if (!data.id) throw new Error('LogicPill: data.id is undefined');
  const defaultHandleData = {
    nodeId: data.id,
    nodeName: data.name,
    nodeType: data.type,
  };

  const onInputUpdated = (value: string, input: InputNode, idx: number) => {
    let logicNode = { ...(graphology.getNodeAttributes(node.id) as LogicNodeAttributes) };
    if (!logicNode) throw new Error('LogicPill: logicNode is undefined');
    let logicNodeInputs = { ...logicNode.inputs };
    if (logicNodeInputs[input.name] != value) {
      logicNodeInputs[input.name] = value;
      logicNode.inputs = logicNodeInputs;
      graphology.setNodeAttribute<any>(node.id, 'inputs', logicNodeInputs); // FIXME: I'm not sure why TS requires <any> to work here
      dispatch(setQuerybuilderNodes(graphology.export()));
    }
  };

  const createLeftHandles = useCallback(
    (sideInputs: InputNode[], positionSide: Position, handleType: HandleType) => {
      let numOfInputs = 0;
      let ret = sideInputs.map((input, i) => {
        let inputTextBox = null;
        if (
          !connectionsToLeft.some(
            (edge) =>
              edge?.attributes?.targetHandleData.nodeId === data.id && edge?.attributes?.targetHandleData.attributeName === input.name
          )
        ) {
          inputTextBox = (
            <Input
              className={styles.logicInput}
              // value={0}
              style={{ top: -5, transform: `translateY(-${i * 20}%)` }}
              // onChange={(e) => onInputUpdated(e.target.value, input)}
              onKeyDown={(e) => {
                if (e.key === 'Enter') onInputUpdated((e.target as HTMLInputElement).value, input, i);
              }}
              onBlur={(e) => onInputUpdated(e.target.value, input, i)}
            />
          );
          numOfInputs++;
        }
        return (
          <Handle
            type={handleType}
            position={positionSide}
            id={toHandleId({ ...defaultHandleData, attributeName: input.name, attributeType: input.type })} // TODO
            key={input.name + input.type}
            // style={{ top: `${((i + 0.8) / (side.length + 0.6)) * 120}%` }}
            style={{ top: `${((i + 0.8) / (sideInputs.length + 0.6)) * 100}%` }}
            className={styleHandleMap[input.type]}
          >
            {inputTextBox}
          </Handle>
        );
      });
      return { handles: ret, number: numOfInputs };
    },
    [node]
  );
  const { handles: leftHandles, number: leftInputsNumber } = createLeftHandles(node.data.logic.inputs, Position.Left, 'target');

  return (
    <div className={styles.logic}>
      <span className={styles.logicSpan} style={{ height: `${Math.max(leftInputsNumber * 2.6, 1.5)}rem` }}>
        {
          <span>
            {connectionsToLeft.map((e) => e?.attributes?.sourceHandleData.attributeName)}.{output.name}
          </span>
        }
        {leftHandles}
        {!!node.data.logic.output && (
          <Handle
            type={'source'}
            position={Position.Right}
            id={toHandleId({ ...defaultHandleData, attributeName: output.name, attributeType: output.type })} // TODO
            // style={{ top: `${((i + 0.8) / (side.length + 0.6)) * 100}%` }}
            className={styleHandleMap?.[output.type]}
          ></Handle>
        )}
      </span>
    </div>
  );
}
